import java.util.*;

//***************************************************************************

class Buffer{

  Vector buffer;
  boolean linemode;
  Vector oob_buffer;      // The out of band buffer (non-linemode only)

  String unfinished;      // Stuff not done to be put at the end
  String unfinished_top;  // Stuff not done to be put at the front

  boolean stopped = false;

//---------------------------------------------------------------------------

  Buffer(){
    this( false );
  }

//---------------------------------------------------------------------------

  Buffer( boolean lm ){
    linemode = lm;
    buffer     = new Vector();
    oob_buffer = new Vector();
    unfinished     = null;
    unfinished_top = null;
    stopped = false;
  }


//---------------------------------------------------------------------------

  public synchronized void putOOB( String s ){
    // The OOB buffer is always in non-linemode

    oob_buffer.addElement( s );
    notify();
  }

//---------------------------------------------------------------------------

  public synchronized void put( String s ){

    String st;
    String line;

    int startstring;
    int endstring;

    //System.out.println( "Enter put for " + s.length() + " bytes" );

    // Add any unfinished text from a previous buffer to the beginning
    if( unfinished != null ){
      st = unfinished + s;
      unfinished = null;
    } else {
      st = s;
    }

    if( !linemode ){

      // Not in linemode, so add the entire string at once
      buffer.addElement( st );

    } else {

      // Break it up by newlines, store them, and store the rest in
      // the "unfinished" buffer

      startstring = 0;   // Start at the beginning of the string
      do{
        endstring = st.indexOf( '\n', startstring );
	if( endstring != -1 ){
	  try{
	    line = st.substring( startstring, endstring+1 );
	    buffer.addElement( line );
	    startstring = endstring + 1;
	  } catch( StringIndexOutOfBoundsException e ){
	  }
	}
      } while( endstring != -1 );

      // Add the rest
      unfinished = st.substring( startstring );
    }
    notify();
    //System.out.println( "Leaving put" );
  }

//---------------------------------------------------------------------------

  public synchronized void push( String s ){
    // Put this at the top of the stack

    String st;
    String line;

    int startstring;
    int endstring;

    //System.out.println( "Enter push for " + s.length() + " bytes" );

    // Add any unfinished text from a previous push to the beginning
    if( unfinished_top != null ){
      st = unfinished_top + s;
      unfinished_top = null;
    } else {
      st = s;
    }

    if( !linemode ){

      // Not in linemode, so add the entire string at once
      buffer.insertElementAt( st, 0 );

    } else {

      // Break it up by newlines, store them, and store the rest in
      // the "unfinished_top" buffer

      startstring = 0;   // Start at the beginning of the string
      do{
        endstring = st.indexOf( '\n', startstring );
	if( endstring != -1 ){
	  try{
	    line = st.substring( startstring, endstring+1 );
	    buffer.insertElementAt( line, 0 );
	    startstring = endstring + 1;
	  } catch( StringIndexOutOfBoundsException e ){
	  }
	}
      } while( endstring != -1 );

      // Add the rest
      unfinished_top = st.substring( startstring );
    }
    notify();
    //System.out.println( "Leaving push" );
  }

//---------------------------------------------------------------------------

  public synchronized String get(){
    //System.out.println( "Enter get" );

    String ret;

    while( buffer.isEmpty() && oob_buffer.isEmpty() && !stopped ){
      try {
        wait();
      } catch (InterruptedException e) {
        System.out.println( "InterruptedException in buffer" );
      }
    }

    if( !oob_buffer.isEmpty() ){    
      ret = oob_buffer.firstElement().toString();
      oob_buffer.removeElementAt( 0 );
    } else if( !buffer.isEmpty() ){
      ret = buffer.firstElement().toString();
      buffer.removeElementAt( 0 );
    } else {
      ret = "";
    }

    //System.out.println( "Leaving get for " + ret.length() + " bytes" );
    return ret;
  }

//---------------------------------------------------------------------------

  public synchronized void stop(){
    System.out.println( "buffer stopping" );
    stopped = true;
    notify();
    System.out.println( "buffer stopped" );
  }

//---------------------------------------------------------------------------

  public boolean isLinemode(){
    return linemode;
  }

//---------------------------------------------------------------------------

  public void setLinemode( boolean lm ){
    linemode = lm;
  }

//---------------------------------------------------------------------------

  public int size(){
    return sizePlain() + sizeOOB();
  }

//---------------------------------------------------------------------------

  public int sizePlain(){
    return buffer.size();
  }

//---------------------------------------------------------------------------

  public int sizeOOB(){
    return oob_buffer.size();
  }

//---------------------------------------------------------------------------

  public boolean isOOB(){
    return !oob_buffer.isEmpty();
  }

//---------------------------------------------------------------------------

}
