import java.awt.*;
import java.util.*;
import java.net.*;
import java.io.*;
import java.applet.Applet;

import CalInNet;
import CalOutNet;
import Parse;

//***************************************************************************

public class Cal extends Applet {

  boolean started;
  static boolean isApplet = true;

  Buffer tonet;
  Buffer fromnet;
  Buffer tofirstwindow;

  Socket network;

  CalAbort  abort;
  CalStatus status;

  // Host information to connect to
  // String defaulthost = "lily.acm.rpi.edu";
  String defaulthost = "eclipse.its.rpi.edu";
  // String defaulthost = "128.113.24.33";
  int    defaultport = 7777;
  String host = null;
  int    port = 0;

  CalOutNet netout;
  CalInNet  netin;
  Parse     parser;
  CalSystem system;

//---------------------------------------------------------------------------

  public void init(){
    started = false;

    // Draw the graphic
    CalGraphic graphic = new CalGraphic();

    abort = new CalAbort();
    status = new CalStatus();
    status.add( "Starting Cafe au Lys" );
    Panel statusbar = new Panel();
    statusbar.setLayout( new BorderLayout() );
    statusbar.add( "Center", status );
    statusbar.add( "West", abort );

    setLayout( new BorderLayout() );
    add( "Center", graphic );
    add( "South", statusbar );
    show();
    paint( getGraphics() );
  
  }

//---------------------------------------------------------------------------

  public void start(){

    if( !started ){

      if( isApplet ){
        CalValue.put( "SYSTEM", "apptype", "applet" );
        host = getParameter( "host" );
        Integer port_int  = Integer.getInteger( "port" );
        if( port_int != null ){
          port = port_int.intValue();
        }
      } else {
        CalValue.put( "SYSTEM", "apptype", "application" );
        try{
          host = CalValue.get( "SYSTEM", "host" );
        } catch( CalUnknownValueException cuve ){
        }
      }

      CalValue.put( "SYSTEM", "status", "starting" );

      system = new CalSystem( status, this );
      system.start();

      if( host == null ){
        host = defaulthost;
      }
      if( port == 0 ){
        port = defaultport;
      }
      status.add( "Connecting to " + host + ":" + port + "..." );
 
      try {
        network = new Socket( host, port );
        status.add( "Connected to " + host + ":" + port + "." );
        tonet   = new Buffer();
        fromnet = new Buffer();
  
        netout = new CalOutNet( status, tonet, network );
        netin  = new CalInNet( status, fromnet, network );

	parser = new Parse( status, fromnet, tonet );

  
        // Start the threads
        netout.start();
        netin.start();
	parser.start();

        started = true;
        CalValue.put( "SYSTEM", "status", "running" );

      } catch( UnknownHostException uhe ){
        status.add( "Unknown host " + host );
      } catch( IOException ioe ){
        status.add( "IO exception connecting to " + host );
      }

      try{
        system.join();   // Don't close up till the system thread is done
      } catch( InterruptedException ie ){
      }
  
    }

    stop();

  }  

//---------------------------------------------------------------------------

//  public void close_up(){
  public void stop(){

    System.out.println( "closing up" );
//    try{
//      system.join();   // Don't close up till the system thread is done
//    } catch( InterruptedException ie ){
//    }

    // Turn off the "bail out" button
    abort.enable( false );

    // Stop the major threads
    parser.stop();
    netout.stop();
    netin.stop();

    // Close the net
    try{
      network.close();
    } catch( IOException e ){
      status.add( "Can't close network: " + e.getMessage() );
    }

    System.out.println( "Close_up" );
  }

//---------------------------------------------------------------------------

  public static void main(String args[]) {

    // @@@ Quick hack to support a single command line option
    try{
      if( args[0].equalsIgnoreCase( "-h" ) ){
        CalValue.put( "SYSTEM", "host", args[1] );
      }
    } catch( ArrayIndexOutOfBoundsException aioobe ){
    }


    Frame f = new Frame("Cafe au Lys");
    Cal program = new Cal();

    isApplet = false;

    //Font oldFont = program.getFont();
    //if (oldFont == null) {
    //  System.out.println("Eek! font is null!");
    //} else {
    //  program.setFont(new Font(oldFont.getFamily(), oldFont.getStyle(),
	   //                    oldFont.getSize()+2));
    //}

    program.init();

    f.add("Center", program);
    f.resize(500, 250);
    f.pack();
    f.show();

    program.start();
  }
//---------------------------------------------------------------------------

  public Dimension minimumSize() {
    return new Dimension(500, 250);
  }

//---------------------------------------------------------------------------

  public Dimension preferredSize() {
    return minimumSize();
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalGraphic extends Canvas{

//---------------------------------------------------------------------------

  CalGraphic(){
  }

//---------------------------------------------------------------------------

  public void paint( Graphics g ){

    // Draw the mug
    g.drawRect( 100, 100, 100, 100 );
    g.drawArc( 180, 130, 40, 40, 270, 180 );
    g.drawArc( 170, 120, 60, 60, 270, 180 );
 
    // Draw the lily
    g.drawLine( 150, 10, 150, 100 );
    g.drawArc( 125, 25, 25, 50, 270, 180 );
    g.drawArc( 150, 25, 25, 50, 270, -180 );

    // And some text
    g.drawString( "Cafe au Lys - 0.3.8 alpha", 250, 50 );

  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalStatusLine{

  Date   time;
  String desc;

//---------------------------------------------------------------------------

  CalStatusLine( Date d, String s ){
    time = d;
    desc = s;
  }

//---------------------------------------------------------------------------

  CalStatusLine( String s ){
    this( new Date(), s );
  }

//---------------------------------------------------------------------------

  public Date getTime(){
    return time;
  }

//---------------------------------------------------------------------------

  public String getDesc(){
    return desc;
  }

//---------------------------------------------------------------------------

  public String getString(){
    return( '(' + time.toLocaleString() + ") - " + desc );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalStatus extends Panel{

  TextField field;
  Vector events;

//---------------------------------------------------------------------------

  CalStatus(){
    field = new TextField();
    field.setEditable( false );
    setLayout( new BorderLayout() );
    add( "Center", field );
    add( "East", new Button( "Status" ) );

    events = new Vector();
  }

//---------------------------------------------------------------------------

  public void add( String message ){
    field.setText( message );
    events.addElement( new CalStatusLine( message ) );
  }

//---------------------------------------------------------------------------

  public boolean handleEvent( Event e ){
    if( e.target instanceof Button ){

      // This is a status notice
      add( "Opening CaL Status window" );

      // Create a window showing events to date
      new CalStatusWindow( this, events );
      
      return true;
    } 
    return false;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalStatusWindow extends Frame{

  CalStatus status;
  TextArea area;

//---------------------------------------------------------------------------

  CalStatusWindow( CalStatus s, Vector v ){

    status = s;

    area = new TextArea();
    area.setEditable( false );

    setLayout( new BorderLayout() );
    add( "Center", area );
    add( "South", new Button( "Close" ) );
    pack();
    show();

    CalStatusLine stat_line;
    String line;
    int co = 0;
    try{
      while( true ){
        stat_line = (CalStatusLine)v.elementAt( co );
        line = stat_line.getString();
        area.appendText( line + '\n' );
        co++;
      }
    } catch( ArrayIndexOutOfBoundsException e ){
      // Normal - we're out of the loop
    }
  }

//---------------------------------------------------------------------------

  public boolean handleEvent( Event e ){
    if( e.target instanceof Button ){

      // Close ourselves
      dispose();

      // This is a status notice
      status.add( "Closing CaL Status window" );

      return true;
    } 
    return false;
  }

//---------------------------------------------------------------------------

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalAbort extends Button{

//---------------------------------------------------------------------------

  CalAbort(){
    super( "Bail Out" );
  }

//---------------------------------------------------------------------------

  public boolean handleEvent( Event e ){
    if( e.target instanceof CalAbort ){

      // Set our terminate status
      CalValue.put( "SYSTEM", "status", "shutdown" );

      // Get out
      return true;
    } 
    return false;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalSystem implements Runnable{

  Thread thread;

  CalStatus statusbar;
  Cal applet;

//---------------------------------------------------------------------------

  CalSystem( CalStatus s, Cal a ){
    statusbar = s;
    applet = a;
  }

//---------------------------------------------------------------------------

  public void start(){
    if( thread == null ){
      thread = new Thread( this, "CalSystem" );
      thread.start();
    }
  }

//---------------------------------------------------------------------------

  public void run(){
    String oldstatus = "";
    String status;
    try{
      status = CalValue.get( "SYSTEM", "status" );
    } catch( CalUnknownValueException uve1 ){
      status = "";
    }
    
    while( !status.equalsIgnoreCase( "shutdown" ) ){
      try{
        thread.sleep( 1000 );
        status = CalValue.get( "SYSTEM", "status" );
      } catch( InterruptedException ie ){
        System.out.println( "InterruptedException in CalSystem" );
      } catch( CalUnknownValueException uve2 ){
        status = "unknown";
      }
      if( status != oldstatus ){
        statusbar.add( "Status changed to '" + status +"'" );
        oldstatus = status;
      }
    }

    try{
      if( CalValue.get( "SYSTEM", "apptype" ).equalsIgnoreCase("applet") ){
        //applet.close_up();
        statusbar.add( "You may now close this page." ); 
        //applet.destroy();   // @@@ Does this work?
      } else {
        // application
        System.exit( 0 );
      }
    } catch( CalUnknownValueException uve3 ){
      statusbar.add( "Unknown app type.  Cannot close down." );
    }

  }

//---------------------------------------------------------------------------

  public void stop(){
    System.out.println( "Stopping " + thread.getName() );
    thread.stop();
    thread = null;
    System.out.println( "Stopped" );
  }

//---------------------------------------------------------------------------

  public void join() throws InterruptedException{
    thread.join();
  }

//---------------------------------------------------------------------------

}

//***************************************************************************
