import java.awt.*;
import java.util.*;

import CalWindow;

//***************************************************************************

class CalMenu extends MenuBar{

//---------------------------------------------------------------------------

  CalMenu( CalWindow parent ){
    add( new MenuFile( parent ) );
    add( new MenuList( parent ) );
    add( new MenuWindow( parent ) );

    MenuHelp help = new MenuHelp( parent );
    add( help );
    setHelpMenu( help );
  }

//---------------------------------------------------------------------------

  public boolean postEvent( Event evt ){
    if( evt.target instanceof MenuItemAction ){
      return ((MenuItemAction)evt.target).action( evt, new Object() );
      // @@@ new Object   is a hack!
    }

    return super.postEvent( evt );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

interface MenuItemAction{
  public abstract boolean action( Event evt, Object obj );
}

//***************************************************************************

class MenuList extends Menu{

  public static String LIST = "list...";

//---------------------------------------------------------------------------

  MenuList( CalWindow parent ){
    super( "List", true );    // Allow tear-offs

    Buffer      parser  = parent.getParseBuffer();

    Menu whoMenu = new Menu( "Who" );    
    whoMenu.add( new MenuListWho( "Here", parser ) );
    whoMenu.add( new MenuListWho( "Away", parser ) );
    whoMenu.add( new MenuListWho( "Detached", parser ) );
    whoMenu.add( new MenuListWho( "All", parser ) );
    whoMenu.add( new MenuListWho( "Everyone", parser ) );
    whoMenu.addSeparator();
    whoMenu.add( new MenuListWho( LIST, parser ) );

    Menu whatMenu = new Menu( "What" );
    whatMenu.add( new MenuListWhat( "Active", parser ) );
    whatMenu.add( new MenuListWhat( "Inactive", parser ) );
    whatMenu.add( new MenuListWhat( "All", parser ) );
    whatMenu.addSeparator();
    whatMenu.add( new MenuListWhat( LIST, parser ) );

    add( whoMenu );
    add( whatMenu );
    
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuWindow extends Menu{

//---------------------------------------------------------------------------

  MenuWindow( CalWindow parent ){
    super( "Window", true );    // Allow tear-offs

    Buffer      parser  = parent.getParseBuffer();
    Buffer      network = parent.getNetworkBuffer();
    String      name    = parent.getName();

    LineCounter lines   = null;
    CalArea     area    = null;
    
    if( parent instanceof CalTextWindow ){
      lines = ((CalTextWindow)parent).getLines();
      area  = ((CalTextWindow)parent).getArea();
    }

    add( new MenuWindowPause( lines ) );
    add( new MenuWindowResetpauselines( lines, area ) );
    addSeparator();
    add( new MenuWindowClose( name, parser) );
    
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuFile extends Menu{

//---------------------------------------------------------------------------

  MenuFile( CalWindow parent ){
    super( "File", true );    // Allow tear-offs
    
    Buffer      parser  = parent.getParseBuffer();

    Menu quitMenu = new Menu( "Disconnect" );
    quitMenu.add( new MenuFileDetach( parser ) );
    quitMenu.add( new MenuFileBye( parser ) );

    add( new MenuFileLog() );
    addSeparator();
    add( quitMenu );
    add( new MenuFileQuit( parser ) );
    
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuHelp extends Menu{

//---------------------------------------------------------------------------

  MenuHelp( CalWindow parent ){
    super( "Help", true );    // Allow tear-offs
    
    add( new MenuHelpNewuser() );
    add( new MenuHelpSubmit() );
    addSeparator();
    add( new MenuHelpAboutCal() );
    add( new MenuHelpAboutLily() );
    
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuFileLog extends CheckboxMenuItem implements MenuItemAction{

//---------------------------------------------------------------------------

  MenuFileLog(){
    super( "Log..." );
    enable( false );       // @@@ Implement this function
  }  

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    return false;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuWindowPause extends CheckboxMenuItem implements MenuItemAction{

  LineCounter lines;

//---------------------------------------------------------------------------

  MenuWindowPause( LineCounter l ){
    super( "Pause" );
    lines = l;
    setState( true );      // We start out pausing
    if( l != null ){
      enable( true );
    } else {
      enable( false );
    }
  }  

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    boolean state = getState();
    lines.enable( state );
    return true;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuWindowResetpauselines extends MenuItem implements MenuItemAction{

  LineCounter lines;
  CalArea area;

//---------------------------------------------------------------------------

  MenuWindowResetpauselines( LineCounter l, CalArea a ){
    super( "Pause reset" );
    lines = l;
    area = a;
    if( (l != null) && (a != null) ){
      enable( true );
    } else {
      enable( false );
    }
  }  

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    area.charLines();
    return true;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuFileDetach extends MenuItem implements MenuItemAction{

  Buffer parser;

//---------------------------------------------------------------------------

  MenuFileDetach( Buffer p ){
    super( "Detach..." );
    enable( true );
    parser = p;
  }  

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    parser.put( "%cal disconnect detach\n" );
    return true;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuFileBye extends MenuItem implements MenuItemAction{

  Buffer parser;

//---------------------------------------------------------------------------

  MenuFileBye( Buffer p ){
    super( "Bye..." );
    enable( true );
    parser = p;
  }  

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    parser.put( "%cal disconnect bye\n" );
    return true;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuFileQuit extends MenuItem implements MenuItemAction{

  Buffer parser;

//---------------------------------------------------------------------------

  MenuFileQuit( Buffer p ){
    super( "Quit..." );
    enable( true );
    parser = p;
  }  

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    parser.put( "%cal quit\n" );
    return true;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuWindowClose extends MenuItem implements MenuItemAction{

  Buffer parser;
  String window_name;

//---------------------------------------------------------------------------

  MenuWindowClose( String wn, Buffer p ){
    super( "Close" );
    if( wn.equals( Parse.DEFAULT ) ){
      // Do not allow the default window to be destroyed
      enable( false );
    } else {
      enable( true );
    }
    window_name = wn;
    parser = p;
  }

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    
    parser.put( "%cal close " + window_name + '\n' );
    return true;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuListWho extends MenuItem implements MenuItemAction{

  String command;
  Buffer parser;

//---------------------------------------------------------------------------

  MenuListWho( String cmd, Buffer p ){
    super( cmd );
    enable( true );
    command = cmd;
    parser = p;
  }

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    if( command.equals( MenuList.LIST ) ){
      // They want to enter in the user
    } else {
      parser.put( "%cal wholist " + command + '\n' );
    }
    return true;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuListWhat extends MenuItem implements MenuItemAction{

  String command;
  Buffer parser;

//---------------------------------------------------------------------------

  MenuListWhat( String cmd, Buffer p ){
    super( cmd );
    enable( true );
    command = cmd;
    parser = p;
  }

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    if( command.equals( MenuList.LIST ) ){
      // They want to enter in the discussion
    } else {
      parser.put( "%cal whatlist " + command + '\n' );
    }
    return true;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuHelpNewuser extends MenuItem implements MenuItemAction{

//---------------------------------------------------------------------------

  MenuHelpNewuser(){
    super( "New Users..." );
    enable( false );       // @@@ Implement this function
  }  

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    return false;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuHelpSubmit extends MenuItem implements MenuItemAction{

//---------------------------------------------------------------------------

  MenuHelpSubmit(){
    super( "Submit a Bug Report" );
    enable( false );       // @@@ Implement this function
  }  

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    return false;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuHelpAboutCal extends MenuItem implements MenuItemAction{

//---------------------------------------------------------------------------

  MenuHelpAboutCal(){
    super( "About CaL..." );
    enable( false );       // @@@ Implement this function
  }  

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    return false;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class MenuHelpAboutLily extends MenuItem implements MenuItemAction{

//---------------------------------------------------------------------------

  MenuHelpAboutLily(){
    super( "About lily..." );
    enable( false );       // @@@ Implement this function
  }  

//---------------------------------------------------------------------------

  public boolean action( Event evt, Object obj ){
    return false;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************
