import java.awt.*;
import java.util.*;

import CalTextWindow;
import Buffer;

//***************************************************************************

class CalUnknownBuffException extends RuntimeException{

//---------------------------------------------------------------------------

  CalUnknownBuffException(){
    super();
  }

//---------------------------------------------------------------------------

  CalUnknownBuffException( String s ){
    super( s );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class Parse implements Runnable{

  Thread thread;

  CalStatus status;

  Buffer fromnet;  // What we get from the server
  Buffer tonet;    // If we have to talk to the server

  BufferTable buffers;
  static final String DEFAULT = "Default";
  String redirect = null;            // Where to redirect output to

  boolean logged_in = false;
  boolean reviewing = false;
  boolean connected = false;
  String uls = ""; // Unknown login stuff    @@@

  final String idString = "lily 0.8b";
  //final String idString = "CaL 0.3";

//---------------------------------------------------------------------------

  Parse( CalStatus s, Buffer from, Buffer to ){
    status = s;
    fromnet = from;
    tonet   = to;

    buffers = new BufferTable();
  }

//---------------------------------------------------------------------------

  public void start(){
    if( thread == null ){
      thread = new Thread( this, "Parser" );
      thread.start();
    }
  }

//---------------------------------------------------------------------------

  String detached = "do you wish to review now? (Y/n)";

  public void run(){

    while( true ){
      String text = fromnet.get();
      if( !logged_in ){
        status.add( "Logging in" );
        text = parse_login( text );
      } 

      if( logged_in ){
        if( !connected ){

          if( text.indexOf( detached ) != -1 ){
            status.add( "Attaching" );
            tonet.put( "n\n" );   // Do not ask for the review
          }

          if( text.indexOf( "%connected" ) != -1 ){
            connected = true;
            fromnet.setLinemode( true );
            status.add( "Fully connected" );

            // Bring up the initial frame
            Buffer defbuffer = new Buffer();
            CalTextWindow defaultwindow = 
              new CalTextWindow( DEFAULT, status, defbuffer, tonet, fromnet );
            defbuffer.setLinemode( true );
            buffers.put( DEFAULT, defbuffer );

            // Send the text to the new buffer
            // This may include the connected attach message - it will
            //  process
            try{
              send_to_buffer( DEFAULT, text );
              send_to_buffer( DEFAULT, "\n" );
            } catch( CalUnknownBuffException e ){
              status.add( "Unknown Buffer: " + e.getMessage() );
            }

          }

        } else if( text.indexOf( "%" ) == 0 ){
          parse_percent( text );
        } else if( text.indexOf( "(" ) == 0 ){
          parse_status( text );
        } else {
          // When there is noplace else to go...
          default_send( text );
        }
      }
    }
  }

//---------------------------------------------------------------------------

  void default_send( String text ){

    String name;
    if( redirect != null ){
      name = redirect;
    } else {
      name = DEFAULT;
    }

    try{
      send_to_buffer( name, text );
    } catch( CalUnknownBuffException e ){
      status.add( "Unknown Buffer: " + e.getMessage() );
    }
  }

//---------------------------------------------------------------------------

  void oob_default_send( String text ){

    String name;
    if( redirect != null ){
      name = redirect;
    } else {
      name = DEFAULT;
    }

    try{
      oob_send_to_buffer( name, text );
    } catch( CalUnknownBuffException e ){
      status.add( "Unknown Buffer: " + e.getMessage() );
    }
  }

//---------------------------------------------------------------------------

  String parse_login( String line ){

    boolean parsed = false;
    // Each if statement is seperate, since it is possible for 
    // more than one part to show up in the same packet or parsable
    // set of packets.

    System.out.println( "parse_login: '" + line + "'" );
    uls = uls + line;

    if( uls.indexOf( "login:" ) != -1 ){
      // We are being prompted for login
      LoginBox login = new LoginBox( uls, tonet );
      parsed = true;
    } 

    if( uls.indexOf( "*** Timed-out waiting for login. ***" ) != -1 ){
      // Close it up if we lost the connection
      CalValue.put( "SYSTEM", "status", "shutdown" );
    }

    if( (uls.indexOf("*** Connected ***") != -1) ||
        (uls.indexOf("*** Redirecting old connection to this port ***") != -1)
      ){
      // Send the client ID once connected
      tonet.put( "#$# client " + idString + '\n' );
      parsed = true;
    } 

    if( (uls.indexOf( "Please choose a name" ) != -1) &&
        (uls.indexOf( "-->" ) != -1 ) ){
      NameBox name = new NameBox( uls, tonet );
      parsed = true;
    }

    if( (uls.indexOf( "blurb" ) != -1) &&
        (uls.indexOf( "-->" ) != -1 ) ){
      BlurbBox blurb = new BlurbBox( uls, tonet );
      parsed = true;
    }

    if( uls.indexOf( "type /HELP for an introduction" ) != -1 ){
      // @@@ Need to find a better marker, but see the connected
      //     boolean too
      status.add( "Logged in" );
      logged_in = true;
    } 

    if( !parsed ){
      System.out.println( "Unknown login stuff: " + line );
    } else {
      // It was parsed, so clear the uls
      uls = "";
    }

    return uls;
  }

//---------------------------------------------------------------------------

  void parse_percent( String command_line ){
    String command = null;
    String parameter[];
    int num_param;

    parameter = CalUtil.split( command_line );
    command = parameter[0];

    // Now we can take action
    if( command.equalsIgnoreCase( "%cal" ) ){
      parse_percent_cal( parameter );
    } else if( command.equalsIgnoreCase( "%sender" ) ){
      CalValue.put( "GLOBAL", "last_recip", parameter[1] );
    } else if( command.equalsIgnoreCase( "%sendgroup" ) ){
      CalValue.put( "GLOBAL", "last_group", parameter[1] );
    } else if( command.indexOf( "%g" ) == 0 ){
      // This one should be last, since it is not checking the full
      // command word, but the first part of the line

      // Send the signal to the window to handle
      oob_default_send( "%g\n" );

      // Then send the rest of the line there
      try{
        default_send( command_line.substring( 2 ) );
      } catch( StringIndexOutOfBoundsException e ){
      }

    } else {
      status.add( "Unknown % command '" + command + "'" );
    }

  }

//---------------------------------------------------------------------------

  void parse_percent_cal( String param[] ){
    String command = param[1];   // Not 0, since thats "%cal"

    if( command.equalsIgnoreCase( "close" ) ){
      try{
        oob_send_to_buffer( param[2], "%cal close\n" );
        buffers.remove( param[2] );   // Take it out of our list
        status.add( "Removed window " + param[2] );
      } catch( CalUnknownBuffException e ){
        status.add( "Can't delete unknown buffer: " + e.getMessage() );
      }
    } else if( command.equalsIgnoreCase( "wholist" ) ){
      String name = buffers.uniqueName( "who " + param[2] );
      Buffer whobuffer = new Buffer();
      whobuffer.setLinemode( true );
      CalWhoWindow whowindow = 
        new CalWhoWindow( name, status, whobuffer, tonet, fromnet, param[2] );
      buffers.put( name, whobuffer );
      status.add( "Creating new who list: " + name );
    } else if( command.equalsIgnoreCase( "whatlist" ) ){
      String name = buffers.uniqueName( "what " + param[2] );
      Buffer whatbuffer = new Buffer();
      whatbuffer.setLinemode( true );
      CalWhatWindow whatwindow = 
        new CalWhatWindow(name, status, whatbuffer, tonet, fromnet, param[2]);
      buffers.put( name, whatbuffer );
      status.add( "Creating new what list: " + name );
    } else if( command.equalsIgnoreCase( "review" ) ){
      // @@@ Should probably setup for more than just the detach buffer
      String name = buffers.uniqueName( "Detach" );
      Buffer reviewbuffer = new Buffer();
      reviewbuffer.setLinemode( true );
      CalTextWindow window = 
        new CalTextWindow( name, status, reviewbuffer, tonet, fromnet );
      buffers.put( name, reviewbuffer );
      status.add( "Creating new review window: " + name );
      tonet.put( "#$# delimiters " + 
                 "\"%cal redirect "+name+"\" " +
                 "\"%cal nodirect "+name+"\"\n" +
                 "/REVIEW DETACH\n"
               );
    } else if( command.equalsIgnoreCase( "command" ) ){
      // Create a new window and direct the output of the command
      // to there without further parsing
      String commd = CalUtil.unsplit( param, " ", 2 );
      String name = buffers.uniqueName( "Command " + commd );
      Buffer buffer = new Buffer();
      buffer.setLinemode( true );
      CalTextWindow window = 
        new CalTextWindow( name, status, buffer, tonet, fromnet );
      buffers.put( name, buffer );
      status.add( "Creating new command window: " + name );
      tonet.put( "#$# delimiters " + 
                 "\"%cal redirect "+name+"\" " +
                 "\"%cal nodirect "+name+"\"\n" +
                 commd + "\n"
               );
    } else if( command.equalsIgnoreCase( "redirect" ) ){
      try{
        send_to_buffer( param[2], CalUtil.unsplit( param, " " ) + '\n' );
        redirect = param[2];
      } catch( CalUnknownBuffException e ){
        status.add( "Can't redirect to unknown buffer: " +  e.getMessage() );
      }
    } else if( command.equalsIgnoreCase( "nodirect" ) ){
      try{
        send_to_buffer( redirect, CalUtil.unsplit( param, " " ) + '\n' );
      } catch( CalUnknownBuffException e ){
        status.add( "Can't undirect to unknown buffer: " +  e.getMessage() );
      }
      tonet.put( "#$# delimiters\n" );
      redirect = null;
    } else if( command.equalsIgnoreCase( "quit" ) ){
      // Bring up the window to shut down CaL
      AreYouSureQuitWindow quitWindow = new AreYouSureQuitWindow();
    } else if( command.equalsIgnoreCase( "disconnect" ) ){
      AreYouSureDisconnectWindow disconnectWindow = 
        new AreYouSureDisconnectWindow( tonet, param[2] );
    } else {
      status.add( "Unknown %cal command '" + command + "'" );
    }    
  }

//---------------------------------------------------------------------------

  void parse_status( String text ){
    if( text.indexOf( "(thank you for using lily)" ) == 0 ){
      try{
        if( CalValue.get( "SYSTEM", "status" ).equalsIgnoreCase( "standby" ) ){
          CalValue.put( "SYSTEM", "status", "shutdown" );
          System.out.println( "go to shutdown" );
        }
      } catch( CalUnknownValueException uve ){
        // Put us in shutdown state anyway
        status.add( "System status unknown.  Shutting down anyway." );
        CalValue.put( "SYSTEM", "status", "shutdown" );
      }
      // @@@ Should signal the net is about to die and handle
      default_send( text );    // @@@ Send to all?
    } else {
      default_send( text );
    }
  }

//---------------------------------------------------------------------------

  void send_to_buffer( String name, String text ) 
  throws CalUnknownBuffException{
    Buffer tobuf = (Buffer)buffers.get( name );
    if( tobuf != null ){
      tobuf.put( text );
    } else {
      throw new CalUnknownBuffException( name );
    }
  }

//---------------------------------------------------------------------------

  void oob_send_to_buffer( String name, String text ) 
  throws CalUnknownBuffException{
    Buffer tobuf = (Buffer)buffers.get( name );
    if( tobuf != null ){
      tobuf.putOOB( text );
    } else {
      throw new CalUnknownBuffException( name );
    }
  }

//---------------------------------------------------------------------------

  public void stop(){

    System.out.println( "Stopping " + thread.getName() );
    // Tell all of our buffers to die
    Enumeration buffer_list = buffers.keys();
    while( buffer_list.hasMoreElements() ){
      String buffer_name = (String)buffer_list.nextElement();
      parse_percent( "%cal close " + buffer_name );      
    }

    // And stop the thread
    thread.stop();
    thread = null;
    System.out.println( "STOPPED" );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class LoginBox extends Frame{

  String message;
  Buffer tonet;

  Button ok;
  TabTextField loginField;
  TabTextField passwordField;
  Checkbox     doScript;

//---------------------------------------------------------------------------

  LoginBox( String m, Buffer to ){
    super( "CaL Login" );

    tonet = to;

    // Set the message to what lily prompts us
    try{
      int loc = m.lastIndexOf( "login:" );
      message = m.replace( '\r', ' ' )
                 .replace( '\n', ' ' )
                 .substring( 0, loc )
                 .trim();
    } catch( StringIndexOutOfBoundsException e ){
      message = "Welcome";
    }

    ok = new Button( "Ok" );
    passwordField = new TabTextField( 10, ok );
    passwordField.setEchoCharacter( '*' );
    loginField = new TabTextField( 10, passwordField );

    Panel login = new Panel();
    login.setLayout( new BorderLayout() );
    login.add( "West", new Label( "Login: " ) );
    login.add( "East", loginField );

    Panel password = new Panel();
    password.setLayout( new BorderLayout() );
    password.add( "West", new Label( "Password: " ) );
    password.add( "East", passwordField );
 
    doScript = new Checkbox( "Execute startup script" );
    doScript.setState( true );
    doScript.enable( false );    // @@@ This feature isn't implemented

    Panel dialog = new Panel();
    dialog.setLayout( new BorderLayout() );
    dialog.add( "North", login );
    dialog.add( "Center", password );
    dialog.add( "South", doScript );

    setLayout( new BorderLayout() );
    add( "North", new Label( message, Label.CENTER ) );
    add( "Center", dialog );
    add( "South", ok );
    pack();
    show();
    
  }

//---------------------------------------------------------------------------

  public boolean handleEvent( Event evt ){
    if( evt.target instanceof Button ){
      // They pressed the button

      // Get the strings, and if they are blank, put in obvious garbage
      String login = loginField.getText();
      if( login.length() == 0 ){
        login = "x";
      }
      String pass  = passwordField.getText();
      if( pass.length() == 0 ){
        pass = "x";
      }
      
      // Get the data from the logins and send it
      tonet.put( login + ' ' + pass + '\n' );

      // Destroy ourselves
      dispose();
      return true;
    }

    // Other events
    return false;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class NameBox extends Frame{

  Buffer tonet;

  List names;

//---------------------------------------------------------------------------

  NameBox( String data, Buffer to ){
    super( "CaL Reserved Names" );
    tonet = to;

    TextArea motd = new TextArea( 10, 80 );
    motd.setEditable( false );
    Button ok = new Button( "OK" );
    names = new List( 3, false );

    StringTokenizer dataLines = new StringTokenizer( data, "\n\r", false );
    while( dataLines.hasMoreTokens() ){
      String line = dataLines.nextToken();
      if( line.indexOf( '*' ) == 0 ){
        // Part of the MOTD
        motd.appendText( line + '\n' );
      } else if( line.indexOf( "last login at" ) == 1 ){
        // Put the last login info into the MOTD
        motd.appendText( '\n' + line + '\n' );
      } else if( line.indexOf( ">>>>" ) == 1 ){
        int first_quote = 0;
        int last_quote = 0;
        // This line contains the login names
        while( first_quote != -1 ){
          System.out.println( "First quote at " + first_quote );
          first_quote = line.indexOf( '"', last_quote+1 );
          if( first_quote != -1 ){
            last_quote = line.indexOf( '"', first_quote+1 );
            try{
              String name = line.substring( first_quote+1, last_quote );
              names.addItem( name );
            } catch( StringIndexOutOfBoundsException e ){
            }
          }
          
        }
      } else {
        System.out.println( "No clue: " + line );
      }
    }
    
    setLayout( new BorderLayout() );
    add( "North", motd );    
    add( "East", ok );
    add( "Center", names );
    pack();
    show();

  }

//---------------------------------------------------------------------------

  public boolean handleEvent( Event evt ){

    String name;

    if( evt.target instanceof Button ){
      // They pressed the button
      
      // Get the data from the logins and send it
      name = names.getSelectedItem();
      if( name == null ){
        name = names.getItem( 0 );
      }
      tonet.put( name + '\n' );      

      // Destroy ourselves
      dispose();
      return true;
    }

    // Other events
    return false;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class BlurbBox extends Frame{

  Buffer tonet;

  TabTextField blurb;

//---------------------------------------------------------------------------

  BlurbBox( String data, Buffer to ){
    super( "CaL Blurb Entry" );
    tonet = to;

    TextArea motd = new TextArea( 10,80 );
    motd.setEditable( false );
    Button ok = new Button ( "OK" );
    blurb = new TabTextField( 40, ok );

    StringTokenizer dataLines = new StringTokenizer( data, "\n\r", false );
    while( dataLines.hasMoreTokens() ){
      String line = dataLines.nextToken();
      if( line.indexOf( '*' ) == 0 ){
        // Part of the MOTD
        motd.appendText( line + '\n' );
      } else if( line.indexOf( "last login at" ) == 1 ){
        // Put the last login info into the MOTD
        motd.appendText( '\n' + line + '\n' );
      } else if( line.indexOf( "blurb" ) != -1 ){
        motd.appendText( '\n' + line + '\n' );
      } else {
        System.out.println( "No clue: " + line );
      }
    }
    
    setLayout( new BorderLayout() );
    add( "North", motd );
    add( "Center", blurb );
    add( "South", ok );
    pack();
    show();    
  }

//---------------------------------------------------------------------------

  public boolean handleEvent( Event evt ){

    String name;

    if( evt.target instanceof Button ){
      // They pressed the button
      
      // Get the data from the logins and send it
      tonet.put( blurb.getText() + '\n' );      

      // Destroy ourselves
      dispose();
      return true;
    }

    // Other events
    return false;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class TabTextField extends TextField{

  Component next;

//---------------------------------------------------------------------------  

  TabTextField( Component n ){
    super();
    next = n;
  }

//---------------------------------------------------------------------------  

  TabTextField( int cols, Component n ){
    super( cols );
    next = n;
  }

//---------------------------------------------------------------------------  

  TabTextField( String text, Component n ){
    super( text );
    next = n;
  }

//---------------------------------------------------------------------------  

  TabTextField( String text, int cols, Component n ){
    super( text, cols );
    next = n;
  }

//---------------------------------------------------------------------------  

  public boolean handleEvent( Event evt ){

    if( evt.key == '\t' ){
      if( evt.id == Event.KEY_RELEASE ){
        next.requestFocus();
        return true;
      } else {
        // Swalllow other tab events
        return true;
      }
    }

    return false;
  }

//---------------------------------------------------------------------------  

}

//***************************************************************************

class BufferTable extends Hashtable{

//---------------------------------------------------------------------------

  public BufferTable( int initialCapacity, float loadFactor ){
    super( initialCapacity, loadFactor );
  }

//---------------------------------------------------------------------------

  public BufferTable( int initialCapacity ){
    super( initialCapacity );
  }

//---------------------------------------------------------------------------

  public BufferTable( ){
    super( );
  }

//---------------------------------------------------------------------------

  public String uniqueName( String s ){
    // Take the base name and turn it into a "standard", removing duplicates
    //  and spaces and other apparent nasties
    String ret;
    String basename;
    int co = 1;
    
    // Turn the spaces into underscores and the quotes into splats
    basename = s.replace( ' ', '_' )
                .replace( '"', '*' )
                .replace( '\'', '*' );
    ret = basename;

    while( containsKey( ret ) ){
      ret = basename + "<" + co + ">";
      co++;
    }

    return ret;    
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

