import java.awt.*;
import java.net.*;
import java.util.*;

import Buffer;

//***************************************************************************

abstract class CalWindow extends Frame implements Runnable{

  String name;
  CalStatus status;
  Buffer parser;
  Buffer network;
  Buffer input;

  Thread thread;
  boolean running;

//---------------------------------------------------------------------------

  CalWindow( String n, CalStatus stat, 
             Buffer fromnet, Buffer tonet, Buffer toparser ){
    super( "CaL: " + n );

    name = n;
    input = fromnet;
    parser = toparser;
    network = tonet;
    status = stat;

    System.out.println( toString() );
    System.out.println( super.toString() );
  }

//---------------------------------------------------------------------------

  public boolean handleEvent( Event evt ){
  // @@@ was action

    if( evt.id == Event.WINDOW_DESTROY ){
      return true;
//    } else if( evt.target instanceof MenuItemAction ){
//      return ((MenuItemAction)evt.target).action( evt, new Object() );
    } else if( evt.id == Event.GOT_FOCUS ){
      System.out.println( "CalWindow got focus" );
      return super.handleEvent( evt );
    }

    return super.handleEvent( evt );
  }

//---------------------------------------------------------------------------

  public void start(){
    if( thread == null ){
      thread = new Thread( this, name );
      thread.start();
    }
  }

//---------------------------------------------------------------------------

  abstract public void run();
  // This needs to be a loop, generally centering around a 
  //   while( running ){}
  // block with a
  //   dispose()
  // at the end
  // @@@ Should the dispose be in the stop method?

//---------------------------------------------------------------------------

  public void stop(){
    System.out.println( "Stopping " + thread.getName() );
    if( thread != null ){
      thread.stop();
      thread = null;
    }
    System.out.println( "stopped" );
  }

//---------------------------------------------------------------------------

  public String getName(){
    return name;
  }

//---------------------------------------------------------------------------

  public Buffer getParseBuffer(){
    return parser;
  }

//---------------------------------------------------------------------------

  public Buffer getNetworkBuffer(){
    return network;
  }

//---------------------------------------------------------------------------

  public Buffer getInputBuffer(){
    return input;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalTextWindow extends CalWindow{

  LineCounter linecounter;

  CalArea        area;
  CalField       field;
  //CalPauseButton pause;
  Button         pause;
  CalMenu        pullDowns;

  Buffer fromnet;
  Buffer tonet;
  Buffer parser;

//---------------------------------------------------------------------------

  CalTextWindow( String n, CalStatus st,
                 Buffer fn, Buffer tn, Buffer tp ){
    super( n, st, fn, tn, tp );

    fromnet = fn;
    tonet = tn;
    parser = tp;

    linecounter = new LineCounter( 24 );

    area  = new CalArea( linecounter );
    field = new CalField( tonet, fromnet, linecounter );
    //pause = new CalPauseButton( "More", linecounter, fromnet, area,
    //                            this, parser );
    pause = new Button( "More" );
    pullDowns = new CalMenu( this );

    // Do layout
    Panel entryPanel = new Panel();
    entryPanel.setLayout( new BorderLayout() );
    entryPanel.add( "Center", field );
    entryPanel.add( "East", pause );

    setLayout( new BorderLayout() );
    setMenuBar( pullDowns );
    add( "Center", area );
    add( "South", entryPanel );
    pack();
    show();

//    // Start the threads
//    pause.start();
    start();

//    // Block on terminating until the thread stops
//    try{
//      pause.join();      // Stop the thread
//    } catch( InterruptedException e ){
//      System.out.println( "Interrupted Exception" );
//    }
//    System.out.println( "Dispose 2" );
//    super.dispose();
//    System.out.println( "Dispose 3" );

  }

//---------------------------------------------------------------------------

  public void run(){
    running = true;

    while( running ){
      int size = fromnet.size();

      while( (size > 0) && (running) ){
        if( !linecounter.isMaxed() || fromnet.isOOB() ){
          // We can get data either if there is OOB data existing, or
          // we haven't run out of room to print stuff

          // Get the string from the buffer, blocking
          String newstring = fromnet.get();
          if( newstring.indexOf( '%' ) == 0 ){
            percent_command( newstring );
          } else {
            String leftover = area.storeText( newstring );
            if( leftover != null ){
              fromnet.push( leftover );
            }
          }

        } else {
          try{
            thread.sleep( 1000 );
          } catch( InterruptedException e ){
          }
        }
        size = fromnet.size();
        pause.setLabel( String.valueOf( size ) );
        pause.enable();
      } 
      // There is either nothing in the buffer, or we're closing up

      pause.setLabel( "More" );
      pause.disable();

      if( running ){
        // If we are still going, wait a second so we yield time
        try{
          thread.sleep( 1000 );
        } catch( InterruptedException e ){
        }
      }

    } // End while running 
    System.out.println( "not running" );
    dispose();
  }

//---------------------------------------------------------------------------

  void percent_command( String line ){

    String parameter[] = CalUtil.split( line );
    String command = parameter[0];

    if( command.equalsIgnoreCase( "%cal" ) ){
      if( parameter[1].equalsIgnoreCase( "close" ) ){
        System.out.println( "got close" );
        running = false;
        // buff.stop();
      } else if( parameter[1].equalsIgnoreCase( "redirect" ) ){
        // Don't do anything
      } else if( parameter[1].equalsIgnoreCase( "nodirect" ) ){
        // Don't do anything
      } else {
        fromnet.push( "Unknown command: '" + line + "'\n" );
      }
    } else if( command.equalsIgnoreCase( "%recip_regexp" ) ){
      // Ignore
    } else if( command.equalsIgnoreCase( "%g" ) ){
      // Send a signal 
      System.err.print( '\007' );     // @@@ Need real audio
      System.err.flush();
    } else if( command.equalsIgnoreCase( "%connected" ) ){
      if( parameter[1].equalsIgnoreCase( "attach" ) ){
        // Create a review buffer
        parser.put( "%cal review\n" );
      }
    } else {
      fromnet.push( "Unknown command: '" + line + "'\n" );
    }

  }

//---------------------------------------------------------------------------

  public boolean handleEvent( Event evt ){
  // @@@ was action

    if( evt.id == Event.KEY_ACTION_RELEASE ){
      // @@@ None of this works either, either here or in handleEvent
      // Function keys?
      if( evt.key == Event.F11 ){
        // @@@ more dorking
        field.requestFocus();
        return true;
      } else if( evt.key == Event.F12 ){
        // @@@ done because the above is dorked
        // Trigger the MORE button
        linecounter.clear();
      
        // Get out
        return true;
      }
    } else if( evt.id == Event.KEY_PRESS ){
      // @@@ None of this works either, either here or in handleEvent
      // Transfer focus to the entry area.
      // @@@ This translate doesn't quite do what you think
      field.requestFocus();
      evt.translate( field.location().x, field.location().y );
      return false;
    } else if( evt.target == pause ){
      // Resete the line counter
      linecounter.clear();
      return true;
    }

    return super.handleEvent( evt );
  }

//---------------------------------------------------------------------------

  CalArea getArea(){
    return area;
  }

//---------------------------------------------------------------------------

  CalField getField(){
    return field;
  }

//---------------------------------------------------------------------------

  LineCounter getLines(){
    return linecounter;
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalArea extends Panel{

  LineCounter lines;

  Vector text;        // The text entered

  GrCalAreaCanvas canvas;
  GrCalScrollbar right;

//---------------------------------------------------------------------------

  CalArea( LineCounter l ){
    lines = l;

    text = new Vector();  // Create a place to store all our text

    // Create the work areas and draw
    canvas = new GrCalAreaCanvas( text );
    right = new GrCalScrollbar( Scrollbar.VERTICAL, 0, 0, 0, text.size() );
    setLayout( new BorderLayout() );
    add( "Center", canvas );
    add( "East", right );

  }

//---------------------------------------------------------------------------

  public boolean handleEvent(Event evt) {
    if (evt.target instanceof Scrollbar) {
      int val = ((Scrollbar)evt.target).getValue();
      canvas.setFirstLine( val );
      repaint();
      return true;
    } 
    return super.handleEvent( evt );;
  }


//---------------------------------------------------------------------------

  public String storeText( String s ){
    // Attempt to print the text, return any portion that could not
    // be printed or null if all printed

    String ret = null;

    int maxLineWidth = 80;    // @@@ Automatically compute

    // Ensure resize correctly
    charLines();

    // Get rid of unsightly characters  (this is cheating)
    //String ns = s.replace( '\n', ' ' ).replace( '\r', ' ' );
    String ns = CalUtil.remove( s, "\n\r" );

    // Add it to the list of lines, one line at a time
    int pos = 0;
    boolean done = false;
    boolean leftover = true;
    while( !lines.isMaxed() && !done ){
      try{
        text.addElement( ns.substring( pos, pos+maxLineWidth ) );
      } catch( StringIndexOutOfBoundsException e ){
        // There was more than maxLineWidth characters - we're at the end
        try{
          text.addElement( ns.substring( pos ) );
        } catch( StringIndexOutOfBoundsException ei ){
        }
        done = true;
        leftover = false;
      }
      lines.inc();
      pos += maxLineWidth;
    }

    // If there was stuff not printed, send it back
    if( leftover ){
      ret = ns.substring( pos );
    }

    // Set the scrollbar
    int size = text.size();
    right.setMaximum( size );
    right.setValue( size );

    // Put it on the canvas
    canvas.setLastLine( size );

    return ret;
  }

//---------------------------------------------------------------------------

  public void charLines(){

    // Load the data from the text portion
    int num = canvas.numLines();

    // Set some values in the scrollbar
    right.setVisible( num );
    
    // Then save it for the overflow line counter
    lines.setMax( num );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class GrCalAreaCanvas extends Canvas{

  Vector text;
  int firstline;

  Font font;

//---------------------------------------------------------------------------

  GrCalAreaCanvas( Vector t ){
    text = t;
    firstline = 0;

    // This won't work since you don't have a toolkit here???
    //String fonts[] = getToolkit().getFontList();
    //int cof;
    //for( cof=0; cof<fonts.length; cof++ ){
    //  System.out.println( fonts[cof] );
    //}

    font = new Font( "Courier", Font.PLAIN, 12 );
    setFont( font );
  }

//---------------------------------------------------------------------------

  public void paint( Graphics g ){

    int numLines = numLines();
    int fontHeight   = getFontMetrics( getFont() ).getHeight();
    int y = fontHeight;
    int co;
    String string;

    for( co=firstline; co<firstline+numLines; co++ ){

      try{
        string = (String)text.elementAt( co );
        g.drawString( string, 0, y );
        y += fontHeight;
      } catch( ArrayIndexOutOfBoundsException e ){
      }      

    }    
  }

//---------------------------------------------------------------------------

  public void setFirstLine( int l ){
    firstline = l;
    repaint();
  }

//---------------------------------------------------------------------------

  public void setLastLine( int l ){
    int num = numLines();
    if( l < num ){
      setFirstLine( 0 );
    } else {
      setFirstLine( l - num );
    }
  }

//---------------------------------------------------------------------------

  public int numLines(){
    int windowHeight = size().height;
    int fontHeight   = getFontMetrics( getFont() ).getHeight();
    int num = (windowHeight / fontHeight);
    return num;
  }

//---------------------------------------------------------------------------

  public Dimension minimumSize() {
    int fontw = getFontMetrics( font ).charWidth( 'W' );
    int fonth = getFontMetrics( font ).getHeight();
    return new Dimension(80*fontw, 24*fonth);
  }

//---------------------------------------------------------------------------

  public Dimension preferredSize() {
    return minimumSize();
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class GrCalScrollbar extends Scrollbar{

//---------------------------------------------------------------------------

  GrCalScrollbar( int orientation,
                  int value, int visible,
                  int minimum, int maximum ){
    super( orientation, value, visible, minimum, maximum );
  }

//---------------------------------------------------------------------------

  public void setVisible( int visible ){
    setValues( getValue(), visible, getMinimum(), getMaximum() );
  }

//---------------------------------------------------------------------------

  public void setMinimum( int min ){
    setValues( getValue(), getVisible(), min, getMaximum() );
  }

//---------------------------------------------------------------------------

  public void setMaximum( int max ){
    setValues( getValue(), getVisible(), getMinimum(), max );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

//class CalField extends TextField{
class CalField extends TextArea{

  Buffer tonet;
  Buffer fromnet;

  Socket network;

  LineCounter lines;

  Vector sent_to;
  int sent_to_ptr;

  Vector sent_msg;
  int sent_msg_ptr;
  String sent_msg_save;

//---------------------------------------------------------------------------

  CalField( Buffer b, Buffer b2, LineCounter l ){

    super( 2, 80 );

    tonet = b;
    fromnet = b2;

    lines = l;

    sent_to = new Vector();
    sent_to_ptr = -1;

    sent_msg = new Vector();
    sent_msg_ptr = -1;
    sent_msg_save = "";

  }

//---------------------------------------------------------------------------

  boolean insertName( String value ){
    // Only do something if nothing is entered
    //    @@@ should be if its the first
    int l = getText().length();
    if( l == 0 ){
      try{
        // Insert the name and the semicolon at the beginning

        // @@@ It is arguable this shouldn't always be a semicolon,
        // because of how we are doing "sendee" parsing at a newline.

        String name = CalValue.get( "GLOBAL", value );
        int len = name.length();

        insertText( name + ';', 0 );

        return true;
      } catch( CalUnknownValueException e ){
        // Don't do any processing.  No value.
        return false;
      }
    } else {
      // Do not do any processing on it, there is already data
      return false;
    }
  }

//---------------------------------------------------------------------------

  int findEndName(){
    // Try to treat the string as a name and a message, seperated by
    // a semicolon (@@@ use recip_regexp!).  The name cannot have
    // spaces.
    //
    // Return -1 if there is no name, otherwise return the index
    // of the ;
    
    int ret = -1;
    String text = getText();
    int firstspace = text.indexOf( ' ' );
    int firstsemi = text.indexOf( ';' );
    if( (firstsemi != -1) &&
        ( (firstsemi < firstspace) || (firstspace == -1) ) ){
        ret = firstsemi;
    }

    return ret;
  }
  
//---------------------------------------------------------------------------
  
  void replaceName( String name ){
    // Search for a name, if any, and replace it.  If there is no
    // name, just insert one.
    // Name should have trailing ;

    int semicolon = findEndName();
    if( semicolon != -1 ){
      // There was a name
      replaceText( name, 0, semicolon+1 );
    } else {
      // No name already there.  Insert.
      insertText( name, 0 );
    }
  }

//---------------------------------------------------------------------------

  void scrollMsg( int dir ){

    System.out.println( "scrollMsg " + dir );

    String newtext = "**error**";
  
    if( sent_msg_ptr == -1 ){
      sent_msg_save = getText();
    }

    sent_msg_ptr += dir;

    if( sent_msg_ptr == -2 ){
      // Looped bakcwards
      sent_msg_ptr = sent_msg.size() - 1;
    }

    try{
      newtext = (String)sent_msg.elementAt( sent_msg_ptr );
    } catch( ArrayIndexOutOfBoundsException except ){
      newtext = sent_msg_save;
      sent_msg_ptr = -1;
    }

    setText( newtext );
  }

//---------------------------------------------------------------------------

  void handleEnter(){

    // Get the value in the entry area
    String text = getText() + '\n';
    //String text = getText();

    // Clear the line counter if we're not maxed out
    // Do this before we add the text to the buffer so the message
    //   itself can't max out the buffer
    if( !lines.isMaxed() ){
      lines.clear();
    }

    // Send it to the network and the return buffer
    tonet.put( text );
    fromnet.put( text );

    // Try to determine who we just sent it to
    // @@@ Should be using recip-regexp here!
    int firstsemi = findEndName();
    if( firstsemi != -1 ){
      String last_send = text.substring( 0, firstsemi );

      // Set this as our last send
      CalValue.put( "GLOBAL", "last_send", last_send );
       // Put it in the list of people we've sent to
      sent_to.insertElementAt( last_send, 0 );
    }

    // Store the message
    sent_msg.insertElementAt( text, 0 );

    // We might have been using the sent lists, so reset them to the beginning
    sent_to_ptr = -1;
    sent_msg_ptr = -1;

    // Clear the entry
    setText( "" );
  }

//---------------------------------------------------------------------------

  public boolean handleEvent( Event e ){
    //System.out.println( "Mod: " + e.modifiers );
    //System.out.println( "key: " + e.key + " F12: " + Event.F12 );
    //System.out.println( "id: " +e.id+ " KAR: " + Event.KEY_ACTION_RELEASE );

    if( e.id == Event.KEY_PRESS ){
      if( e.key == '\n' ){
        handleEnter();
        return true;
      } else if( e.key == ';' ){
        return insertName( "last_send" );
      } else if( e.key == ':' ){
        return insertName( "last_recip" );
      } else if( e.key == '=' ){
        return insertName( "last_group" );
      } else if( e.key == '\t' ){
        if( e.shiftDown() ){
          // Shifted.  Go backwards
          sent_to_ptr--;
          if( sent_to_ptr == -2 ){
            // Looped backwards
            sent_to_ptr = sent_to.size() - 1;
          }
        } else {
          // No shift.  Advance
          sent_to_ptr++;
        }
        try{
          String name = (String)sent_to.elementAt( sent_to_ptr );
          replaceName( name + ';' );
        } catch( ArrayIndexOutOfBoundsException except ){
          replaceName( "" );
          sent_to_ptr = -1;
        }
        return true;
      }

    } else if( e.id == Event.KEY_RELEASE ){
      // Normal keys?
      if( e.metaDown() && (e.key == 'm') ){
        // @@@ This is dorked too
        // Trigger the MORE button
        lines.clear();
      
        // Get out
        return true;
      }
    } else if( e.id == Event.KEY_ACTION_RELEASE ){
      // Function keys?
      if( e.key == Event.F12 ){
        // @@@ done because the above is dorked
        // Trigger the MORE button
        lines.clear();
      
        // Get out
        return true;
      } else if( e.key == Event.UP ){
        // Move to the last message in the list
        scrollMsg( 1 );
      } else if( e.key == Event.DOWN ){
        // Move to the next message in the list
        scrollMsg( -1 );
      }

      // Cannot get it to recognize "normal" keys in this section,
      // they are swallowed up by the text entry area.

    }
    return super.handleEvent( e );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************
//
//class CalPauseButton extends Button implements Runnable{
//
//  LineCounter lines;
//  Buffer buff;
//  CalArea area;
//  CalTextWindow parent;
//  Buffer parser;
//
//  Thread buttonThread;
//
//  boolean running = true;
//
////---------------------------------------------------------------------------
//
//  CalPauseButton( String label, LineCounter l, Buffer b, CalArea a, 
//                  CalTextWindow p, Buffer parse ){
//    setLabel( label );
//    lines = l;
//    buff = b;
//    area = a;
//    parent = p;
//    parser = parse;
//  }
//
////---------------------------------------------------------------------------
//
//  public void start(){
//    if( buttonThread == null ){
//      buttonThread = new Thread( this, "Button" );
//      buttonThread.start();
//    }
//  }
//
////---------------------------------------------------------------------------
//
//  public void run(){
//    while( running ){
//      int size = buff.size();
//
//      while( (size > 0) && (running) ){
//        if( !lines.isMaxed() || buff.isOOB() ){
//          // We can get data either if there is OOB data existing, or
//          // we haven't run out of room to print stuff
//
//          // Get the string from the buffer, blocking
//          String newstring = buff.get();
//          if( newstring.indexOf( '%' ) == 0 ){
//            percent_command( newstring );
//          } else {
//            String leftover = area.storeText( newstring );
//            if( leftover != null ){
//              buff.push( leftover );
//            }
//          }
//
//        } else {
//          try{
//            buttonThread.sleep( 1000 );
//          } catch( InterruptedException e ){
//          }
//        }
//        size = buff.size();
//        setLabel( String.valueOf( size ) );
//        enable();
//      } 
//      // There is either nothing in the buffer, or we're closing up
//
//      setLabel( "More" );
//      disable();
//
//      if( running ){
//        // If we are still going, wait a second so we yield time
//        try{
//          buttonThread.sleep( 1000 );
//        } catch( InterruptedException e ){
//        }
//      }
//
//    } // End while running 
//    System.out.println( "not running" );
//  }
//
////---------------------------------------------------------------------------
//
//  public boolean handleEvent( Event e ){
//    if( e.target instanceof CalPauseButton ){
//
//      // Reset the line counter
//      lines.clear();
//
//      // Put the focus somewhere else...
//      parent.requestFocus();
//
//      // Get out
//      return true;
//    } 
//    return super.handleEvent( e );
//  }
//
////---------------------------------------------------------------------------
//
//  void percent_command( String line ){
//
//    String parameter[] = CalUtil.split( line );
//    String command = parameter[0];
//
//    if( command.equalsIgnoreCase( "%cal" ) ){
//      if( parameter[1].equalsIgnoreCase( "close" ) ){
//        System.out.println( "got close" );
//        running = false;
//        buff.stop();
//      } else if( parameter[1].equalsIgnoreCase( "redirect" ) ){
//        // Don't do anything
//      } else if( parameter[1].equalsIgnoreCase( "nodirect" ) ){
//        // Don't do anything
//      } else {
//        buff.push( "Unknown command: '" + line + "'\n" );
//      }
//    } else if( command.equalsIgnoreCase( "%recip_regexp" ) ){
//      // Ignore
//    } else if( command.equalsIgnoreCase( "%g" ) ){
//      // Send a signal 
//      System.err.print( '\007' );     // @@@ Need real audio
//      System.err.flush();
//    } else if( command.equalsIgnoreCase( "%connected" ) ){
//      if( parameter[1].equalsIgnoreCase( "attach" ) ){
//        // Create a review buffer
//        parser.put( "%cal review\n" );
//      }
//    } else {
//      buff.push( "Unknown command: '" + line + "'\n" );
//    }
//
//  }
//
////---------------------------------------------------------------------------
//
//  public void stop(){
//    System.out.println( "button stop" );
//    if( buttonThread != null ){
//      buttonThread.stop();
//      buttonThread = null;
//    }
//    System.out.println( "stopped" );
//  }
//
////---------------------------------------------------------------------------
//
//  public void join() throws InterruptedException{
//    buttonThread.join();
//  }
//
////---------------------------------------------------------------------------
//
//}
//
//***************************************************************************

class LineCounter{

  int current;
  int max;

  boolean enabled;

//---------------------------------------------------------------------------

  LineCounter( int m ){
    setMax( m );
    clear();
    enabled = true;
  }

//---------------------------------------------------------------------------

  public void clear(){
    current = 0;
  }

//---------------------------------------------------------------------------

  public void inc(){
    current++;
  }

//---------------------------------------------------------------------------

  public boolean isMaxed(){
    if( enabled ){
      return( current >= max );
    } else {
      return false;
    }
  }

//---------------------------------------------------------------------------

  public void setMax( int m ){
    max = m;
  }

//---------------------------------------------------------------------------

  public int getMax(){
    return max;
  }

//---------------------------------------------------------------------------

  public void enable( boolean status ){
    enabled = status;
  }

//---------------------------------------------------------------------------

  public void enable(){
    enable( true );
  }

//---------------------------------------------------------------------------

  public void disable(){
    enable( false );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalListWindowThread implements Runnable{

  // @@@ Most of this thread should be reengineered so there is a 
  //     general purpose scheduler that will execute commands.

  CalListWindow parent;
  CalStatus status;
  String name;
  String command;
  Buffer network;

  Date next_update;   // When do we next update the list

  Thread thread;

//---------------------------------------------------------------------------

  CalListWindowThread( CalListWindow p, CalStatus st, 
                       String n, String c, Buffer tn ){
    parent = p;
    status = st;
    name = n;
    command = c;
    network = tn;
  }

//---------------------------------------------------------------------------

  public void start(){
    if( thread == null ){
      thread = new Thread( this, parent.getName() + " timer" );
      thread.start();
    }
  }

//---------------------------------------------------------------------------

  public void do_command(){
    status.add( "Updating " + name + ": " + command );
    if( !command.equals( MenuList.LIST ) ){
      String order;
      order = "#$# delimiters " + 
              "\"%cal redirect "+name+"\" " +
              "\"%cal nodirect "+name+"\"\n" +
              command + "\n";
      network.put( order );

      // Figre out when we will update the list next
      int update_time = 1;        // @@@ Should have settable time
      next_update = new Date();
      int min = next_update.getMinutes();
      next_update.setMinutes( min + update_time );
    }
  }

//---------------------------------------------------------------------------

  public void run(){
    boolean running = true;

    do_command();
    while( running ){
      try{
        thread.sleep( 1000 );
        parent.do_button_update();   // @@@  Should be in the parents thread
        if( next_update.before( new Date() ) ){
          // We are after the time for updating next
          do_command();
        }
      } catch( InterruptedException e ){
      }
    }

    System.out.println( "done running " + thread.getName() );
  }

//---------------------------------------------------------------------------

  public void stop(){
    System.out.println( "Stopping " + thread.getName() );
    if( thread != null ){
      thread.stop();
      thread = null;
    }
    System.out.println( "stopped" );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalList extends List{

//---------------------------------------------------------------------------

  CalList(){
    super();
  }

//---------------------------------------------------------------------------

  CalList( int rows, boolean multipleSelections ){
    super( rows, multipleSelections );
  }

//---------------------------------------------------------------------------

  public Dimension mimimumSize( int rows ){
    Dimension ret = super.minimumSize( rows );
    ret.width = get_width();

    return ret;
  }

//---------------------------------------------------------------------------

  public Dimension mimimumSize(){
    Dimension ret = super.minimumSize();
    ret.width = get_width();

    return ret;
  }

//---------------------------------------------------------------------------

  public Dimension preferredSize( int rows ){
    Dimension ret = super.preferredSize( rows );
    ret.width = get_width();

    return ret;
  }

//---------------------------------------------------------------------------

  public Dimension preferredSize(){
    Dimension ret = super.preferredSize();
    ret.width = get_width();

    return ret;
  }

//---------------------------------------------------------------------------

  int get_width(){
    int ret;
    ret = getFontMetrics( getFont() ).charWidth( 'W' ) * 80;
    return ret;
  }

//---------------------------------------------------------------------------

  public synchronized void clear(){
    // The List class badly implements this
    // It appears to delete every other item, and the delItem() and
    //    delItems() methods don't throw exceptions

    while( countItems() != 0 ){
      delItem( 0 );
    }
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

abstract class CalListWindow extends CalWindow implements Runnable{

  CalListWindowThread timerthread;
  Thread thread;

  String command;

  Vector strings;
  Vector buttons;

  CalMenu pullDowns;
  CalList listbox;
  Panel buttonbox;

  Font font;

//---------------------------------------------------------------------------

  CalListWindow( String n, CalStatus st,
                 Buffer fromnet, Buffer tonet, Buffer toparser,
                 String c ){
    super( n, st, fromnet, tonet, toparser );

    command = c;

    pullDowns = new CalMenu( this );

    listbox = new CalList( 20, false );
    font = new Font( "Courier", Font.PLAIN, 12 );
    listbox.setFont( font );
    listbox.resize( 500, 200 );

    buttons = new Vector();
    try{
      buttonbox = new Panel();
      buttonbox.setLayout( new GridLayout( 0, 1, 0, 5 ) );
    } catch( IllegalArgumentException e ){ 
    }

    setLayout( new BorderLayout() );
    setMenuBar( pullDowns );
    add( "Center", listbox );
    add( "East", buttonbox );
    pack();
    show();

    // Send the request to the server for our data
    // @@@ Includes the button check function
    timerthread = new CalListWindowThread( this, status, name,
                                           command, tonet );
    timerthread.start();

    // Start the thread that will read and process data to us
    start();
  };

//---------------------------------------------------------------------------

  abstract void parse_line( String line );
  // Take the line and add it to whatever list they maintain

//---------------------------------------------------------------------------

  abstract void create_strings();
  // Take the list of objects and turn it into a list of strings

//---------------------------------------------------------------------------

  void add_button( CalListButton button ){
    buttons.addElement( button );
    buttonbox.add( button );
    pack();
    show();
  }

//---------------------------------------------------------------------------

  void do_button_update(){
    CalListButton button;
    int index = listbox.getSelectedIndex();

    if( index != -1 ){
      int co = 0;
      boolean done = false;
      while( !done ){
        try{
          button = (CalListButton)buttons.elementAt( co );
          button.set_active( index );
          co++;
        } catch( ArrayIndexOutOfBoundsException e ){
          done = true;
        } // End try
      } // End while
    } // End If
  } // End method

//---------------------------------------------------------------------------

  abstract void clear_list();

//---------------------------------------------------------------------------

  void draw_list(){
    strings = new Vector();
    create_strings();
    listbox.clear();
    
    int co = 0;
    boolean done = false;
    String line;
    while( !done ){
      try{
        line = (String)strings.elementAt( co );
        listbox.addItem( CalUtil.remove( line, "\r\n" ) );
        co++;
      } catch( ArrayIndexOutOfBoundsException e ){
        done = true;
      }
    }
  }

//---------------------------------------------------------------------------

  public boolean handleEvent( Event evt ){
  // @@@ was action

    if( evt.target == listbox ){
      return true;
    } else if( evt.target instanceof CalListButton ){
      return ((CalListButton)evt.target).pressed( evt, 
        new Integer( listbox.getSelectedIndex() ) );
    }

    return super.handleEvent( evt );
  }

//---------------------------------------------------------------------------

  public void run(){
    String line;

    running = true;
    while( running ){
      line = input.get();
      //System.out.println( line );
      if( line.indexOf( "%cal nodirect" ) == 0 ){
        // End of the list, so go draw it
        draw_list();
      } else if( line.indexOf( "%cal redirect" ) == 0 ){
        // Start of the list, so we are starting over
        clear_list();
      } else if( line.indexOf( "%cal close" ) == 0 ){
        running = false;
      } else {
        parse_line( line );
      }
    }

    System.out.println( "done running" );
    timerthread.stop();
    dispose();

  }

//---------------------------------------------------------------------------

}

//***************************************************************************

abstract class CalListButton extends Button{

//---------------------------------------------------------------------------

  CalListButton( String n ){
    super( n );
    enable( false );
  }

//---------------------------------------------------------------------------

  abstract boolean pressed( Event evt, Object obj );

//---------------------------------------------------------------------------

  abstract void set_active( int index );

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalCantParseException extends RuntimeException{

//---------------------------------------------------------------------------

  CalCantParseException(){
    super();
  }

//---------------------------------------------------------------------------

  CalCantParseException( String s ){
    super( s );
  }

//---------------------------------------------------------------------------

  CalCantParseException( char c ){
    super( "Character '" + c + "'" );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhoInfoButton extends CalListButton{

  Vector user;
  Buffer parser;

//---------------------------------------------------------------------------

  CalWhoInfoButton( Vector u, Buffer p ){
    super( "Info" );
    user = u;
    parser = p;
  }

//---------------------------------------------------------------------------

  boolean pressed( Event evt, Object obj ){
    int index = ((Integer)obj).intValue();
    String line;
    String nameblurb;
    String namespace;
    String name;
    int blurbindex;

    // @@@ Splitting the name and the blurb is done here because
    // I'm too much of an idiot to split it before

    try{
      line = (String)user.elementAt( index );
      nameblurb = line.substring( 2, 40 );
      blurbindex = nameblurb.indexOf( '[' );
      if( blurbindex != -1 ){
        namespace = nameblurb.substring( 0, blurbindex - 1 );
      } else {
        namespace = nameblurb;
      }
      name = namespace.trim();
      parser.put( "%cal command /INFO \"" + name + "\"\n" );
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
    }
    
    return true;
  }

//---------------------------------------------------------------------------

  void set_active( int index ){
    boolean active = false;
    String line;
    char   indicator;

    try{
      line = (String)user.elementAt( index );
      indicator = line.charAt( 1 );
      if( (indicator == '+') || (indicator == '=') ){
        active = true;
      }
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
   }

    enable( active );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhoFingerButton extends CalListButton{

  Vector user;
  Buffer parser;

//---------------------------------------------------------------------------

  CalWhoFingerButton( Vector u, Buffer p ){
    super( "Finger" );
    user = u;
    parser = p;
  }

//---------------------------------------------------------------------------

  boolean pressed( Event evt, Object obj ){
    int index = ((Integer)obj).intValue();
    String line;
    String nameblurb;
    String namespace;
    String name;
    int blurbindex;

    // @@@ Splitting the name and the blurb is done here because
    // I'm too much of an idiot to split it before

    try{
      line = (String)user.elementAt( index );
      nameblurb = line.substring( 2, 40 );
      blurbindex = nameblurb.indexOf( '[' );
      if( blurbindex != -1 ){
        namespace = nameblurb.substring( 0, blurbindex - 1 );
      } else {
        namespace = nameblurb;
      }
      name = namespace.trim();
      parser.put( "%cal command /FINGER \"" + name + "\"\n" );
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
    }
    
    return true;
  }

//---------------------------------------------------------------------------

  void set_active( int index ){
    boolean active = false;
    String line;
    char   indicator;

    try{
      line = (String)user.elementAt( index );
      indicator = line.charAt( 1 );
      if( (indicator == '-') || (indicator == '=') ){
        active = true;
      }
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
    }

    enable( active );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhoIgnoreButton extends CalListButton{

  Vector user;

//---------------------------------------------------------------------------

  CalWhoIgnoreButton( Vector u ){
    super( "Ignore/Unignore" );
    user = u;
  }

//---------------------------------------------------------------------------

  boolean pressed( Event evt, Object obj ){
    return false;
  }

//---------------------------------------------------------------------------

  void set_active( int index ){
    boolean active = false;
    String  function = "Ignore/Uningnore";
    String line;
    char   indicator;

    try{
      line = (String)user.elementAt( index );
      indicator = line.charAt( 0 );
      if( (indicator == '|') || (indicator == '>') ){
        active = true;
        function = "Unignore";
      } else if( (indicator == ' ') || (indicator == '<') ){
        active = true;
        function = "Ignore";
      } else {
        // I don't know what this character is!
        active = false;
      }
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
    }

    // @@@ This is difficult to implement in reality!
    active = false;

    setLabel( function );
    enable( active );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhoNotifyButton extends CalListButton{

  Vector user;

//---------------------------------------------------------------------------

  CalWhoNotifyButton( Vector u ){
    super( "Notifications..." );
    user = u;
  }

//---------------------------------------------------------------------------

  boolean pressed( Event evt, Object obj ){
    return false;
  }

//---------------------------------------------------------------------------

  void set_active( int index ){
    boolean active = false;
    enable( active );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhoConversationButton extends CalListButton{

  Vector user;

//---------------------------------------------------------------------------

  CalWhoConversationButton( Vector u ){
    super( "Conversation..." );
    user = u;
  }

//---------------------------------------------------------------------------

  boolean pressed( Event evt, Object obj ){
    return false;
  }

//---------------------------------------------------------------------------

  void set_active( int index ){
    boolean active = false;
    String line;
    char   indicator;

    try{
      line = (String)user.elementAt( index );
      indicator = line.charAt( 0 );
      if( indicator == ' ' ){
        // We can only send to them if nobody is ignoring each other
        active = true;
      }
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
    }

    // @@@ Not a valid function yet
    active = false;

    enable( active );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhoWindow extends CalListWindow{

  Vector user;

  Label top;
  Label bottom;

//---------------------------------------------------------------------------

  CalWhoWindow( String n, CalStatus st,
                Buffer fromnet, Buffer tonet, Buffer toparser, 
                String c ){
    super( n, st, fromnet, tonet, toparser, "/WHO " + c );

    user = new Vector();

    add_button( new CalWhoConversationButton( user ) );
    add_button( new CalWhoFingerButton( user, toparser ) );
    add_button( new CalWhoInfoButton( user, toparser ) );
    add_button( new CalWhoNotifyButton( user ) );
    add_button( new CalWhoIgnoreButton( user ) );

    top = new Label();
    top.setFont( font );
    bottom = new Label( "", Label.CENTER );
    add( "North", top );
    add( "South", bottom );    
    pack();
    show();
  }

//---------------------------------------------------------------------------

  void parse_line( String line ){
    // Take the line and add it to whatever list they maintain

    if( line.indexOf( "Name" ) == 2 ){
      // A lame check if this is the first line
      top.setText( CalUtil.remove( line, "\r\n" ) );
    } else if( line.indexOf( "----" ) == 2 ){
      // A lame check if this is the second line
    } else if( line.indexOf( "(there are" ) == 0 ){
      // A lame check if this is the last line
      bottom.setText( CalUtil.remove( line, "\r\n" ) );
    } else {
      // Right now, we just put the string there unparsed
      user.addElement( line );

    }
  }

//---------------------------------------------------------------------------

  void clear_list(){
    user.removeAllElements();
  }

//---------------------------------------------------------------------------

  void create_strings(){
    // Take the list of objects and turn it into a list of strings

    // Since its already a list of strings - copy  @@@
    int co = 0;
    boolean done = false;
    String line;
    while( !done ){
      try{
        line = (String)user.elementAt( co );
        strings.addElement( line );
        co++;
      } catch( ArrayIndexOutOfBoundsException e ){
        done = true;
      }
    }

  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhatJoinButton extends CalListButton{

  Vector disc;
  Buffer parser;

//---------------------------------------------------------------------------

  CalWhatJoinButton( Vector d, Buffer p ){
    super( "Join/Quit" );
    disc = d;
    parser = p;
  }

//---------------------------------------------------------------------------

  boolean ismember( int index ) throws CalCantParseException{
    // Are we a member?
    boolean ret = false;

    String line;
    char   indicator;

    try{
      line = (String)disc.elementAt( index );
      indicator = line.charAt( 0 );
      if( indicator == '*' ){
        ret = true;
      } else if( indicator == ' ' ){
        ret = false;
      } else {
        // What is this?
        throw new CalCantParseException( indicator );
      }
    } catch( ArrayIndexOutOfBoundsException ae ){
      throw new CalCantParseException( "Array OOB" );
    } catch( StringIndexOutOfBoundsException se ){
      throw new CalCantParseException( "String OOB" );
    }

    return ret;
  }

//---------------------------------------------------------------------------

  boolean pressed( Event evt, Object obj ){
    int index = ((Integer)obj).intValue();
    boolean member;
    String line;
    String name;
    String cmd;

    try{
      member = ismember( index );
      if( member ){
        cmd = "/QUIT";
      } else {
        cmd = "/JOIN";
      }

      line = (String)disc.elementAt( index );
      name = line.substring( 2, 12 );
      parser.put( "%cal command " + cmd + ' ' + name + '\n' );
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
    } catch( CalCantParseException ccpe ){
    }
    
    return true;
  }

//---------------------------------------------------------------------------

  void set_active( int index ){
    boolean active = false;
    String function = "Join/Quit";
    boolean member;

    try{
      member = ismember( index );
      if( member ){
        function = "Quit";
        active = true;
      } else {
        function = "Join";
        active = true;
      }
    } catch( CalCantParseException ccpe ){
    }

    setLabel( function );
    enable( active );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhatInfoButton extends CalListButton{

  Vector disc;
  Buffer parser;

//---------------------------------------------------------------------------

  CalWhatInfoButton( Vector d, Buffer p ){
    super( "Info" );
    disc = d;
    parser = p;
  }

//---------------------------------------------------------------------------

  boolean pressed( Event evt, Object obj ){
    int index = ((Integer)obj).intValue();
    boolean member;
    String line;
    String name;

    try{
      line = (String)disc.elementAt( index );
      name = line.substring( 2, 12 );
      parser.put( "%cal command /INFO " + name + '\n' );
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
    }
    
    return true;
  }

//---------------------------------------------------------------------------

  void set_active( int index ){
    boolean active = false;
    String line;
    char   indicator;

    try{
      line = (String)disc.elementAt( index );
      indicator = line.charAt( 1 );
      if( indicator == '+' ){
        active = true;
      }
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
   }

    enable( active );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhatReviewButton extends CalListButton{

  Vector disc;
  Buffer parser;

//---------------------------------------------------------------------------

  CalWhatReviewButton( Vector d, Buffer p ){
    super( "Review" );
    disc = d;
    parser = p;
  }

//---------------------------------------------------------------------------

  boolean pressed( Event evt, Object obj ){
    int index = ((Integer)obj).intValue();
    boolean member;
    String line;
    String name;

    try{
      line = (String)disc.elementAt( index );
      name = line.substring( 2, 12 );
      parser.put( "%cal command /REVIEW " + name + '\n' );
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
    }
    
    return true;

  }

//---------------------------------------------------------------------------

  void set_active( int index ){
    boolean active = false;
    String line;
    char   indicator;

    try{
      line = (String)disc.elementAt( index );
      indicator = line.charAt( 0 );
      if( indicator == '*' ){
        active = true;
      }
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
    }

    enable( active );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhatFingerButton extends CalListButton{

  // This is actually the /WHAT command on a particular discussion
  // but it is so much like /FINGER that... well...

  Vector disc;
  Buffer parser;

//---------------------------------------------------------------------------

  CalWhatFingerButton( Vector d, Buffer p ){
    super( "Finger" );
    disc = d;
    parser = p;
  }

//---------------------------------------------------------------------------

  boolean pressed( Event evt, Object obj ){
    int index = ((Integer)obj).intValue();
    boolean member;
    String line;
    String name;

    try{
      line = (String)disc.elementAt( index );
      name = line.substring( 2, 12 );
      parser.put( "%cal command /WHAT " + name + '\n' );
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
    }
    
    return true;
  }

//---------------------------------------------------------------------------

  void set_active( int index ){
    boolean active = true;
    enable( active );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhatConversationButton extends CalListButton{

  Vector disc;

//---------------------------------------------------------------------------

  CalWhatConversationButton( Vector d ){
    super( "Conversation..." );
    disc = d;
  }

//---------------------------------------------------------------------------

  boolean pressed( Event evt, Object obj ){
    return false;
  }

//---------------------------------------------------------------------------

  void set_active( int index ){
    boolean active = false;

    // @@@ Not installed yet

    enable( active );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhatMembersButton extends CalListButton{

  Vector disc;
  Buffer parser;

//---------------------------------------------------------------------------

  CalWhatMembersButton( Vector d, Buffer p ){
    super( "Members..." );
    disc = d;
    parser = p;
  }

//---------------------------------------------------------------------------

  boolean pressed( Event evt, Object obj ){
    int index = ((Integer)obj).intValue();
    String line;
    String name;

    try{
      line = (String)disc.elementAt( index );
      name = line.substring( 2, 12 );
      parser.put( "%cal wholist " + name + '\n' );
    } catch( ArrayIndexOutOfBoundsException ae ){
    } catch( StringIndexOutOfBoundsException se ){
    }
    
    return true;
  }

//---------------------------------------------------------------------------

  void set_active( int index ){
    boolean active = true;
    enable( active );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhatModerateButton extends CalListButton{

  Vector disc;

//---------------------------------------------------------------------------

  CalWhatModerateButton( Vector d ){
    super( "Moderate..." );
    disc = d;
  }

//---------------------------------------------------------------------------

  boolean pressed( Event evt, Object obj ){
    return false;
  }

//---------------------------------------------------------------------------

  void set_active( int index ){
    boolean active = false;

    // @@@ Not installed yet

    enable( active );
  }

//---------------------------------------------------------------------------

}

//***************************************************************************

class CalWhatWindow extends CalListWindow{

  Vector disc;

  Label top;
  Label bottom;

//---------------------------------------------------------------------------

  CalWhatWindow( String n, CalStatus st,
                 Buffer fromnet, Buffer tonet, Buffer toparser, 
                 String c ){
    super( n, st, fromnet, tonet, toparser, "/WHAT " + c );

    disc = new Vector();

    add_button( new CalWhatConversationButton( disc ) );
    add_button( new CalWhatFingerButton( disc, toparser ) );
    add_button( new CalWhatInfoButton( disc, toparser ) );
    add_button( new CalWhatReviewButton( disc, toparser ) );
    add_button( new CalWhatJoinButton( disc, toparser ) );
    add_button( new CalWhatMembersButton( disc, toparser ) );
    add_button( new CalWhatModerateButton( disc ) );

    top = new Label();
    top.setFont( font );
    bottom = new Label( "", Label.CENTER );
    add( "North", top );
    add( "South", bottom );    
    pack();
    show();
  }

//---------------------------------------------------------------------------

  void parse_line( String line ){
    // Take the line and add it to whatever list they maintain

    if( line.indexOf( "Name" ) == 2 ){
      // A lame check if this is the first line
      top.setText( CalUtil.remove( line, "\r\n" ) );
    } else if( line.indexOf( "----" ) == 2 ){
      // A lame check if this is the second line
    } else if( line.indexOf( "(there are" ) == 0 ){
      // A lame check if this is the last line
      bottom.setText( CalUtil.remove( line, "\r\n" ) );
    } else {
      // Right now, we just put the string there unparsed
      disc.addElement( line );

    }
  }

//---------------------------------------------------------------------------

  void clear_list(){
    disc.removeAllElements();
  }

//---------------------------------------------------------------------------

  void create_strings(){
    // Take the list of objects and turn it into a list of strings

    // Since its already a list of strings - copy  @@@
    int co = 0;
    boolean done = false;
    String line;
    while( !done ){
      try{
        line = (String)disc.elementAt( co );
        strings.addElement( line );
        co++;
      } catch( ArrayIndexOutOfBoundsException e ){
        done = true;
      }
    }

  }

//---------------------------------------------------------------------------

}

//***************************************************************************

